import Resolver, { BasicModuleRegistry, ResolverConfiguration } from '@glimmer/resolver';
import { Dict, ModuleLocator, TemplateLocator, ComponentCapabilities } from '@glimmer/interfaces';
import defaultResolverConfiguration from './default-resolver-configuration';
import { precompile } from './compiler';
import Application, {
  ApplicationConstructor,
  BytecodeLoader,
  DOMBuilder,
  Loader,
  RuntimeCompilerLoader,
  SyncRenderer,
  ModuleTypes,
  INTERNAL_DYNAMIC_SCOPE
} from '@glimmer/application';
import { assert } from '@glimmer/util';
import { BundleCompiler, CompilerDelegate as ICompilerDelegate } from '@glimmer/bundle-compiler';
import { buildAction, mainTemplate, Renderer, BytecodeMetadata } from '@glimmer/application';
import { compilable } from '@glimmer/opcode-compiler';
import { SimpleDocument } from '@simple-dom/interface';
import HTMLSerializer from '@simple-dom/serializer';
import { SSRApplication } from '@glimmer/ssr';

import didRender from './did-render';

export interface AppBuilderOptions<T> {
  appName?: string;
  loader?: string;
  ApplicationClass?: ApplicationConstructor<T>;
  ComponentManager?: any; // TODO - typing
  resolverConfiguration?: ResolverConfiguration;
  document?: SimpleDocument;
}

export interface ComponentFactory {}

export class TestApplication extends Application {
  rootElement: Element;
}

export interface AppBuilderTemplateMeta {
  specifier: string;
}

interface HelperFunction extends Function {
  isFactory?: boolean;
}

interface TestSSRApplicationOptions {
  rootName?: string;
  resolver?: Resolver;
  loader?: Loader;
  renderer?: Renderer;
  serializer?: HTMLSerializer;
  [INTERNAL_DYNAMIC_SCOPE]?: Dict<unknown>;
}

function locatorFor(module: string, name: string): TemplateLocator<ModuleLocator> {
  return {
    kind: 'template',
    module,
    name,
    meta: {
      module,
      name,
    },
  };
}

export class AppBuilder<T extends TestApplication> {
  rootName: string;
  modules: Dict<unknown> = {};
  templates: Dict<string> = {};
  options: AppBuilderOptions<T>;

  constructor(name: string, options: AppBuilderOptions<T>) {
    this.rootName = name;
    this.options = options;
    this.template('Main', '<div />');
    this.helper('action', buildAction, true);
  }

  template(name: string, template: string) {
    assert(
      name.charAt(0) === name.charAt(0).toUpperCase(),
      'template names must start with a capital letter'
    );

    let specifier = `template:/${this.rootName}/components/${name}`;
    this.modules[specifier] = precompile(template, { meta: { specifier } });
    this.templates[specifier] = template;
    return this;
  }

  component(name: string, componentFactory: ComponentFactory) {
    let specifier = `component:/${this.rootName}/components/${name}`;
    this.modules[specifier] = componentFactory;
    return this;
  }

  helper(name: string, helperFunc: HelperFunction, isFactory = false) {
    let specifier = `helper:/${this.rootName}/components/${name}`;
    helperFunc.isFactory = isFactory;
    this.modules[specifier] = helperFunc;
    return this;
  }

  protected buildResolver(): Resolver {
    let resolverConfiguration = this.options.resolverConfiguration || defaultResolverConfiguration;
    resolverConfiguration.app = resolverConfiguration.app || {
      name: this.rootName,
      rootName: this.rootName,
    };

    let registry = new BasicModuleRegistry(this.modules);
    return new Resolver(resolverConfiguration, registry);
  }

  protected buildRuntimeCompilerLoader(): Loader {
    return new RuntimeCompilerLoader();
  }

  protected buildBytecodeLoader(resolver: Resolver) {
    let delegate = new CompilerDelegate(resolver);
    let compiler = new BundleCompiler(delegate);

    let mainLocator = locatorFor('template:mainTemplate', 'default');
    mainLocator.meta.module = 'template:mainTemplate';

    let block = JSON.parse(mainTemplate.block);
    let compilableTemplate = compilable({
      block,
      referrer: mainLocator.meta,
    });

    compiler.addCompilableTemplate(mainLocator, compilableTemplate);

    for (let module in this.templates) {
      compiler.addTemplateSource(locatorFor(module, 'default'), this.templates[module]);
    }

    let { heap, pool, table } = compiler.compile();

    let resolverTable: unknown[] = [];

    let meta: Dict<BytecodeMetadata> = {};

    table.vmHandleByModuleLocator.forEach((vmHandle, locator) => {
      let handle = table.byModuleLocator.get(locator);
      let template = compiler.getTemplate(locator);

      meta[locator.module] = {
        v: vmHandle,
        h: handle,
        sT: template.symbolTable,
      };
    });

    table.byHandle.forEach((locator, handle) => {
      let module = locator.module.replace('template:/', 'component:/');
      if (this.modules[module]) {
        if (module.indexOf('helper:') === 0) {
          const moduleType = (this.modules[module] as HelperFunction).isFactory ? ModuleTypes.HELPER_FACTORY : ModuleTypes.HELPER;
          resolverTable[handle] = [moduleType, this.modules[module]];
        } else {
          resolverTable[handle] = this.modules[module];
        }
      }
    });

    let bytecode = heap.buffer;
    let data = {
      prefix: '',
      mainEntry: table.vmHandleByModuleLocator.get(mainLocator),
      pool,
      table: resolverTable,
      meta,
      heap: {
        table: heap.table,
        handle: heap.handle,
      },
    };

    return new BytecodeLoader({ bytecode, data });
  }

  protected buildLoader(resolver): Loader {
    switch (this.options.loader) {
      case 'runtime-compiler':
        return this.buildRuntimeCompilerLoader();
      case 'bytecode':
        return this.buildBytecodeLoader(resolver);
      default:
        throw new Error(`Unrecognized loader ${this.options.loader}`);
    }
  }

  renderToString(componentName: string, data: Dict<unknown>, options?: TestSSRApplicationOptions): Promise<string> {
    const resolver = this.buildResolver();
    let loader = this.buildLoader(resolver);

    return SSRApplication.renderToString(componentName, data, {
      rootName: this.rootName,
      resolver,
      loader,
      renderer: new SyncRenderer(),
      ...options
    });
  }

  async boot(): Promise<T> {
    let resolver = this.buildResolver();
    let loader: Loader = this.buildLoader(resolver);

    let doc: Document = (this.options.document as Document) || document;
    let element = doc.body;
    let builder = new DOMBuilder({ element });
    let renderer = new SyncRenderer();

    let app = new this.options.ApplicationClass({
      resolver,
      builder,
      loader,
      renderer,
      rootName: this.rootName,
      document: this.options.document,
    });

    let rootElement = doc.createElement('div');
    app.rootElement = rootElement;
    app.renderComponent('Main', rootElement);
    app.boot();

    await didRender(app);

    return app;
  }
}

const CAPABILITIES: ComponentCapabilities = {
  createInstance: true,
  dynamicLayout: false,
  dynamicTag: false,
  wrapped: false,
  prepareArgs: false,
  createArgs: true,
  attributeHook: false,
  elementHook: false,
  updateHook: true,
  createCaller: false,
  dynamicScope: true,
};

class CompilerDelegate implements ICompilerDelegate<ModuleLocator> {
  constructor(protected resolver: Resolver) {}

  hasComponentInScope(name: string, referrer: ModuleLocator): boolean {
    return !!this.resolver.identify(`template:${name}`, referrer.module);
  }

  resolveComponent(name: string, referrer: ModuleLocator): ModuleLocator {
    let resolved = this.resolver.identify(`template:${name}`, referrer.module);
    return { module: resolved, name: 'default' };
  }

  getComponentCapabilities() {
    return CAPABILITIES;
  }

  hasHelperInScope(helperName: string, referrer: ModuleLocator): boolean {
    return !!this.resolver.identify(`helper:${helperName}`, referrer.module);
  }

  resolveHelper(helperName: string, referrer: ModuleLocator): ModuleLocator {
    let resolved = this.resolver.identify(`helper:${helperName}`, referrer.module);
    return { module: resolved, name: 'default' };
  }

  hasPartialInScope(partialName: string, referrer: ModuleLocator): boolean {
    throw new Error('Method not implemented.');
  }

  resolvePartial(partialName: string, referrer: ModuleLocator): ModuleLocator {
    throw new Error('Method not implemented.');
  }

  hasModifierInScope(modifierName: string, referrer: ModuleLocator): boolean {
    throw new Error('Method not implemented.');
  }

  resolveModifier(modifierName: string, referrer: ModuleLocator): ModuleLocator {
    throw new Error('Method not implemented.');
  }
}

function buildApp<T extends TestApplication>(options: AppBuilderOptions<T> = {}): AppBuilder<T> {
  options.appName = options.appName || 'test-app';
  options.loader = options.loader || 'runtime-compiler';
  options.ApplicationClass =
    options.ApplicationClass || (TestApplication as ApplicationConstructor<T>);

  return new AppBuilder(options.appName, options);
}

export { buildApp };
